@@ -465,32 +465,27 @@ struct dir_data {
     rb_encoding *enc;
 };
 
-static void
-dir_mark(void *ptr)
-{
-    struct dir_data *dir = ptr;
-    rb_gc_mark(dir->path);
-}
-
 static void
 dir_free(void *ptr)
 {
     struct dir_data *dir = ptr;
 
     if (dir->dir) closedir(dir->dir);
-    xfree(dir);
 }
 
-static size_t
-dir_memsize(const void *ptr)
-{
-    return sizeof(struct dir_data);
-}
+RUBY_REFERENCES(dir_refs) = {
+    RUBY_REF_EDGE(struct dir_data, path),
+    RUBY_REF_END
+};
 
 static const rb_data_type_t dir_data_type = {
     "dir",
-    {dir_mark, dir_free, dir_memsize,},
-    0, 0, RUBY_TYPED_WB_PROTECTED | RUBY_TYPED_FREE_IMMEDIATELY
+    {
+        RUBY_REFS_LIST_PTR(dir_refs),
+        dir_free,
+        NULL, // Nothing allocated externally, so don't need a memsize function
+    },
+    0, NULL, RUBY_TYPED_WB_PROTECTED | RUBY_TYPED_FREE_IMMEDIATELY | RUBY_TYPED_DECL_MARKING | RUBY_TYPED_EMBEDDABLE
 };
 
 static VALUE dir_close(VALUE);
@@ -591,6 +586,45 @@ dir_s_close(rb_execution_context_t *ec, VALUE klass, VALUE dir)
     return dir_close(dir);
 }
 
+# if defined(HAVE_FDOPENDIR) && defined(HAVE_DIRFD)
+/*
+ * call-seq:
+ *   Dir.for_fd(fd) -> dir
+ *
+ * Returns a new \Dir object representing the directory specified by the given
+ * integer directory file descriptor +fd+:
+ *
+ *   d0 = Dir.new('..')
+ *   d1 = Dir.for_fd(d0.fileno)
+ *
+ * Note that the returned +d1+ does not have an associated path:
+ *
+ *   d0.path # => '..'
+ *   d1.path # => nil
+ *
+ * This method uses the
+ * {fdopendir()}[https://www.man7.org/linux/man-pages/man3/fdopendir.3p.html]
+ * function defined by POSIX 2008;
+ * the method is not implemented on non-POSIX platforms (raises NotImplementedError).
+ */
+static VALUE
+dir_s_for_fd(VALUE klass, VALUE fd)
+{
+    struct dir_data *dp;
+    VALUE dir = TypedData_Make_Struct(klass, struct dir_data, &dir_data_type, dp);
+
+    if (!(dp->dir = fdopendir(NUM2INT(fd)))) {
+        rb_sys_fail("fdopendir");
+        UNREACHABLE_RETURN(Qnil);
+    }
+
+    RB_OBJ_WRITE(dir, &dp->path, Qnil);
+    return dir;
+}
+#else
+#define dir_s_for_fd rb_f_notimplement
+#endif
+
 NORETURN(static void dir_closed(void));
 
 static void
@@ -618,10 +652,13 @@ dir_check(VALUE dir)
 
 
 /*
- *  call-seq:
- *     dir.inspect -> string
+ * call-seq:
+ *   inspect -> string
+ *
+ * Returns a string description of +self+:
+ *
+ *   Dir.new('example').inspect # => "#<Dir:example>"
  *
- *  Return a string describing this Dir object.
  */
 static VALUE
 dir_inspect(VALUE dir)
@@ -655,18 +692,18 @@ dir_inspect(VALUE dir)
 
 #ifdef HAVE_DIRFD
 /*
- *  call-seq:
- *     dir.fileno -> integer
+ * call-seq:
+ *   fileno -> integer
  *
- *  Returns the file descriptor used in <em>dir</em>.
+ * Returns the file descriptor used in <em>dir</em>.
  *
- *     d = Dir.new("..")
- *     d.fileno   #=> 8
- *
- *  This method uses dirfd() function defined by POSIX 2008.
- *  NotImplementedError is raised on other platforms, such as Windows,
- *  which doesn't provide the function.
+ *   d = Dir.new('..')
+ *   d.fileno # => 8
  *
+ * This method uses the
+ * {dirfd()}[https://www.man7.org/linux/man-pages/man3/dirfd.3.html]
+ * function defined by POSIX 2008;
+ * the method is not implemented on non-POSIX platforms (raises NotImplementedError).
  */
 static VALUE
 dir_fileno(VALUE dir)
@@ -685,14 +722,14 @@ dir_fileno(VALUE dir)
 #endif
 
 /*
- *  call-seq:
- *     dir.path -> string or nil
- *     dir.to_path -> string or nil
+ * call-seq:
+ *   path -> string or nil
  *
- *  Returns the path parameter passed to <em>dir</em>'s constructor.
+ * Returns the +dirpath+ string that was used to create +self+
+ * (or +nil+ if created by method Dir.for_fd):
+ *
+ *   Dir.new('example').path # => "example"
  *
- *     d = Dir.new("..")
- *     d.path   #=> ".."
  */
 static VALUE
 dir_path(VALUE dir)
@@ -746,16 +783,18 @@ to_be_skipped(const struct dirent *dp)
 }
 
 /*
- *  call-seq:
- *     dir.read -> string or nil
+ * call-seq:
+ *   read -> string or nil
  *
- *  Reads the next entry from <em>dir</em> and returns it as a string.
- *  Returns <code>nil</code> at the end of the stream.
+ * Reads and returns the next entry name from +self+;
+ * returns +nil+ if at end-of-stream;
+ * see {Dir As Stream-Like}[rdoc-ref:Dir@Dir+As+Stream-Like]:
+ *
+ *   dir = Dir.new('example')
+ *   dir.read # => "."
+ *   dir.read # => ".."
+ *   dir.read # => "config.h"
  *
- *     d = Dir.new("testdir")
- *     d.read   #=> "."
- *     d.read   #=> ".."
- *     d.read   #=> "config.h"
  */
 static VALUE
 dir_read(VALUE dir)
@@ -764,7 +803,7 @@ dir_read(VALUE dir)
     struct dirent *dp;
 
     GetDIR(dir, dirp);
-    errno = 0;
+    rb_errno_set(0);
     if ((dp = READDIR(dirp->dir, dirp->enc)) != NULL) {
         return rb_external_str_new_with_enc(dp->d_name, NAMLEN(dp), dirp->enc);
     }
@@ -784,24 +823,23 @@ dir_yield(VALUE arg, VALUE path)
 }
 
 /*
- *  call-seq:
- *     dir.each { |filename| block }  -> dir
- *     dir.each                       -> an_enumerator
+ * call-seq:
+ *   each {|entry_name| ... } -> self
  *
- *  Calls the block once for each entry in this directory, passing the
- *  filename of each entry as a parameter to the block.
+ * Calls the block with each entry name in +self+:
  *
- *  If no block is given, an enumerator is returned instead.
+ *   Dir.new('example').each {|entry_name| p entry_name }
  *
- *     d = Dir.new("testdir")
- *     d.each  {|x| puts "Got #{x}" }
+ * Output:
+
+ *   "."
+ *   ".."
+ *   "config.h"
+ *   "lib"
+ *   "main.rb"
  *
- *  <em>produces:</em>
+ * With no block given, returns an Enumerator.
  *
- *     Got .
- *     Got ..
- *     Got config.h
- *     Got main.rb
  */
 static VALUE
 dir_each(VALUE dir)
@@ -844,16 +882,17 @@ dir_each_entry(VALUE dir, VALUE (*each)(VALUE, VALUE), VALUE arg, int children_o
 
 #ifdef HAVE_TELLDIR
 /*
- *  call-seq:
- *     dir.pos -> integer
- *     dir.tell -> integer
+ * call-seq:
+ *   tell -> integer
  *
- *  Returns the current position in <em>dir</em>. See also Dir#seek.
+ * Returns the current position of +self+;
+ * see {Dir As Stream-Like}[rdoc-ref:Dir@Dir+As+Stream-Like]:
+ *
+ *   dir = Dir.new('example')
+ *   dir.tell  # => 0
+ *   dir.read  # => "."
+ *   dir.tell  # => 1
  *
- *     d = Dir.new("testdir")
- *     d.tell   #=> 0
- *     d.read   #=> "."
- *     d.tell   #=> 12
  */
 static VALUE
 dir_tell(VALUE dir)
@@ -871,18 +910,24 @@ dir_tell(VALUE dir)
 
 #ifdef HAVE_SEEKDIR
 /*
- *  call-seq:
- *     dir.seek( integer ) -> dir
+ * call-seq:
+ *   seek(position) -> self
  *
- *  Seeks to a particular location in <em>dir</em>. <i>integer</i>
- *  must be a value returned by Dir#tell.
+ * Sets the position in +self+ and returns +self+.
+ * The value of +position+ should have been returned from an earlier call to #tell;
+ * if not, the return values from subsequent calls to #read are unspecified.
+ *
+ * See {Dir As Stream-Like}[rdoc-ref:Dir@Dir+As+Stream-Like].
+ *
+ * Examples:
+ *
+ *   dir = Dir.new('example')
+ *   dir.pos      # => 0
+ *   dir.seek(3)  # => #<Dir:example>
+ *   dir.pos      # => 3
+ *   dir.seek(30) # => #<Dir:example>
+ *   dir.pos      # => 5
  *
- *     d = Dir.new("testdir")   #=> #<Dir:0x401b3c40>
- *     d.read                   #=> "."
- *     i = d.tell               #=> 12
- *     d.read                   #=> ".."
- *     d.seek(i)                #=> #<Dir:0x401b3c40>
- *     d.read                   #=> ".."
  */
 static VALUE
 dir_seek(VALUE dir, VALUE pos)
@@ -900,17 +945,24 @@ dir_seek(VALUE dir, VALUE pos)
 
 #ifdef HAVE_SEEKDIR
 /*
- *  call-seq:
- *     dir.pos = integer  -> integer
+ * call-seq:
+ *   pos = position -> integer
  *
- *  Synonym for Dir#seek, but returns the position parameter.
+ * Sets the position in +self+ and returns +position+.
+ * The value of +position+ should have been returned from an earlier call to #tell;
+ * if not, the return values from subsequent calls to #read are unspecified.
+ *
+ * See {Dir As Stream-Like}[rdoc-ref:Dir@Dir+As+Stream-Like].
+ *
+ * Examples:
+ *
+ *   dir = Dir.new('example')
+ *   dir.pos      # => 0
+ *   dir.pos = 3  # => 3
+ *   dir.pos      # => 3
+ *   dir.pos = 30 # => 30
+ *   dir.pos      # => 5
  *
- *     d = Dir.new("testdir")   #=> #<Dir:0x401b3c40>
- *     d.read                   #=> "."
- *     i = d.pos                #=> 12
- *     d.read                   #=> ".."
- *     d.pos = i                #=> 12
- *     d.read                   #=> ".."
  */
 static VALUE
 dir_set_pos(VALUE dir, VALUE pos)
@@ -923,15 +975,19 @@ dir_set_pos(VALUE dir, VALUE pos)
 #endif
 
 /*
- *  call-seq:
- *     dir.rewind -> dir
+ * call-seq:
+ *   rewind -> self
  *
- *  Repositions <em>dir</em> to the first entry.
+ * Sets the position in +self+ to zero;
+ * see {Dir As Stream-Like}[rdoc-ref:Dir@Dir+As+Stream-Like]:
+ *
+ *   dir = Dir.new('example')
+ *   dir.read    # => "."
+ *   dir.read    # => ".."
+ *   dir.pos     # => 2
+ *   dir.rewind  # => #<Dir:example>
+ *   dir.pos     # => 0
  *
- *     d = Dir.new("testdir")
- *     d.read     #=> "."
- *     d.rewind   #=> #<Dir:0x401b3fb0>
- *     d.read     #=> "."
  */
 static VALUE
 dir_rewind(VALUE dir)
@@ -944,14 +1000,18 @@ dir_rewind(VALUE dir)
 }
 
 /*
- *  call-seq:
- *     dir.close -> nil
+ * call-seq:
+ *   close -> nil
  *
- *  Closes the directory stream.
- *  Calling this method on closed Dir object is ignored since Ruby 2.3.
+ * Closes the stream in +self+, if it is open, and returns +nil+;
+ * ignored if +self+ is already closed:
+ *
+ *   dir = Dir.new('example')
+ *   dir.read     # => "."
+ *   dir.close     # => nil
+ *   dir.close     # => nil
+ *   dir.read # Raises IOError.
  *
- *     d = Dir.new("testdir")
- *     d.close   #=> nil
  */
 static VALUE
 dir_close(VALUE dir)
@@ -975,7 +1035,7 @@ nogvl_chdir(void *ptr)
 }
 
 static void
-dir_chdir(VALUE path)
+dir_chdir0(VALUE path)
 {
     if (chdir(RSTRING_PTR(path)) < 0)
         rb_sys_fail_path(path);
@@ -987,18 +1047,19 @@ static VALUE chdir_thread = Qnil;
 struct chdir_data {
     VALUE old_path, new_path;
     int done;
+    bool yield_path;
 };
 
 static VALUE
 chdir_yield(VALUE v)
 {
     struct chdir_data *args = (void *)v;
-    dir_chdir(args->new_path);
+    dir_chdir0(args->new_path);
     args->done = TRUE;
     chdir_blocking++;
     if (NIL_P(chdir_thread))
         chdir_thread = rb_thread_current();
-    return rb_yield(args->new_path);
+    return args->yield_path ? rb_yield(args->new_path) : rb_yield_values2(0, NULL);
 }
 
 static VALUE
@@ -1009,50 +1070,103 @@ chdir_restore(VALUE v)
         chdir_blocking--;
         if (chdir_blocking == 0)
             chdir_thread = Qnil;
-        dir_chdir(args->old_path);
+        dir_chdir0(args->old_path);
     }
     return Qnil;
 }
 
+static VALUE
+chdir_path(VALUE path, bool yield_path)
+{
+    if (chdir_blocking > 0) {
+        if (rb_thread_current() != chdir_thread)
+            rb_raise(rb_eRuntimeError, "conflicting chdir during another chdir block");
+        if (!rb_block_given_p())
+            rb_warn("conflicting chdir during another chdir block");
+    }
+
+    if (rb_block_given_p()) {
+        struct chdir_data args;
+
+        args.old_path = rb_str_encode_ospath(rb_dir_getwd());
+        args.new_path = path;
+        args.done = FALSE;
+        args.yield_path = yield_path;
+        return rb_ensure(chdir_yield, (VALUE)&args, chdir_restore, (VALUE)&args);
+    }
+    else {
+        char *p = RSTRING_PTR(path);
+        int r = (int)(VALUE)rb_thread_call_without_gvl(nogvl_chdir, p,
+                                                        RUBY_UBF_IO, 0);
+        if (r < 0)
+            rb_sys_fail_path(path);
+    }
+
+    return INT2FIX(0);
+}
+
 /*
- *  call-seq:
- *     Dir.chdir( [ string] ) -> 0
- *     Dir.chdir( [ string] ) {| path | block }  -> anObject
+ * call-seq:
+ *   Dir.chdir(new_dirpath) -> 0
+ *   Dir.chdir -> 0
+ *   Dir.chdir(new_dirpath) {|new_dirpath| ... } -> object
+ *   Dir.chdir {|cur_dirpath| ... } -> object
  *
- *  Changes the current working directory of the process to the given
- *  string. When called without an argument, changes the directory to
- *  the value of the environment variable <code>HOME</code>, or
- *  <code>LOGDIR</code>. SystemCallError (probably Errno::ENOENT) if
- *  the target directory does not exist.
+ * Changes the current working directory.
  *
- *  If a block is given, it is passed the name of the new current
- *  directory, and the block is executed with that as the current
- *  directory. The original working directory is restored when the block
- *  exits. The return value of <code>chdir</code> is the value of the
- *  block. <code>chdir</code> blocks can be nested, but in a
- *  multi-threaded program an error will be raised if a thread attempts
- *  to open a <code>chdir</code> block while another thread has one
- *  open or a call to <code>chdir</code> without a block occurs inside
- *  a block passed to <code>chdir</code> (even in the same thread).
+ * With argument +new_dirpath+ and no block,
+ * changes to the given +dirpath+:
  *
- *     Dir.chdir("/var/spool/mail")
- *     puts Dir.pwd
- *     Dir.chdir("/tmp") do
- *       puts Dir.pwd
- *       Dir.chdir("/usr") do
- *         puts Dir.pwd
- *       end
- *       puts Dir.pwd
+ *   Dir.pwd         # => "/example"
+ *   Dir.chdir('..') # => 0
+ *   Dir.pwd         # => "/"
+ *
+ * With no argument and no block:
+ *
+ * - Changes to the value of environment variable +HOME+ if defined.
+ * - Otherwise changes to the value of environment variable +LOGDIR+ if defined.
+ * - Otherwise makes no change.
+ *
+ * With argument +new_dirpath+ and a block, temporarily changes the working directory:
+ *
+ * - Calls the block with the argument.
+ * - Changes to the given directory.
+ * - Executes the block (yielding the new path).
+ * - Restores the previous working directory.
+ * - Returns the block's return value.
+ *
+ * Example:
+ *
+ *   Dir.chdir('/var/spool/mail')
+ *   Dir.pwd   # => "/var/spool/mail"
+ *   Dir.chdir('/tmp') do
+ *     Dir.pwd # => "/tmp"
+ *   end
+ *   Dir.pwd   # => "/var/spool/mail"
+ *
+ * With no argument and a block,
+ * calls the block with the current working directory (string)
+ * and returns the block's return value.
+ *
+ * Calls to \Dir.chdir with blocks may be nested:
+ *
+ *   Dir.chdir('/var/spool/mail')
+ *   Dir.pwd     # => "/var/spool/mail"
+ *   Dir.chdir('/tmp') do
+ *     Dir.pwd   # => "/tmp"
+ *     Dir.chdir('/usr') do
+ *       Dir.pwd # => "/usr"
  *     end
- *     puts Dir.pwd
+ *     Dir.pwd   # => "/tmp"
+ *   end
+ *   Dir.pwd     # => "/var/spool/mail"
  *
- *  <em>produces:</em>
+ * In a multi-threaded program an error is raised if a thread attempts
+ * to open a +chdir+ block while another thread has one open,
+ * or a call to +chdir+ without a block occurs inside
+ * a block passed to +chdir+ (even in the same thread).
  *
- *     /var/spool/mail
- *     /tmp
- *     /usr
- *     /tmp
- *     /var/spool/mail
+ * Raises an exception if the target directory does not exist.
  */
 static VALUE
 dir_s_chdir(int argc, VALUE *argv, VALUE obj)
@@ -1071,6 +1185,114 @@ dir_s_chdir(int argc, VALUE *argv, VALUE obj)
         path = rb_str_new2(dist);
     }
 
+    return chdir_path(path, true);
+}
+
+#if defined(HAVE_FCHDIR) && defined(HAVE_DIRFD) && HAVE_FCHDIR && HAVE_DIRFD
+static void *
+nogvl_fchdir(void *ptr)
+{
+    const int *fd = ptr;
+
+    return (void *)(VALUE)fchdir(*fd);
+}
+
+static void
+dir_fchdir(int fd)
+{
+    if (fchdir(fd) < 0)
+        rb_sys_fail("fchdir");
+}
+
+struct fchdir_data {
+    VALUE old_dir;
+    int fd;
+    int done;
+};
+
+static VALUE
+fchdir_yield(VALUE v)
+{
+    struct fchdir_data *args = (void *)v;
+    dir_fchdir(args->fd);
+    args->done = TRUE;
+    chdir_blocking++;
+    if (NIL_P(chdir_thread))
+        chdir_thread = rb_thread_current();
+    return rb_yield_values(0);
+}
+
+static VALUE
+fchdir_restore(VALUE v)
+{
+    struct fchdir_data *args = (void *)v;
+    if (args->done) {
+        chdir_blocking--;
+        if (chdir_blocking == 0)
+            chdir_thread = Qnil;
+        dir_fchdir(RB_NUM2INT(dir_fileno(args->old_dir)));
+    }
+    dir_close(args->old_dir);
+    return Qnil;
+}
+
+/*
+ * call-seq:
+ *   Dir.fchdir(fd) -> 0
+ *   Dir.fchdir(fd) { ... } -> object
+ *
+ * Changes the current working directory to the directory
+ * specified by the integer file descriptor +fd+.
+ *
+ * When passing a file descriptor over a UNIX socket or to a child process,
+ * using +fchdir+ instead of +chdir+ avoids the
+ * {time-of-check to time-of-use vulnerability}[https://en.wikipedia.org/wiki/Time-of-check_to_time-of-use]
+ *
+ * With no block, changes to the directory given by +fd+:
+ *
+ *   Dir.chdir('/var/spool/mail')
+ *   Dir.pwd # => "/var/spool/mail"
+ *   dir  = Dir.new('/usr')
+ *   fd = dir.fileno
+ *   Dir.fchdir(fd)
+ *   Dir.pwd # => "/usr"
+ *
+ * With a block, temporarily changes the working directory:
+ *
+ * - Calls the block with the argument.
+ * - Changes to the given directory.
+ * - Executes the block (yields no args).
+ * - Restores the previous working directory.
+ * - Returns the block's return value.
+ *
+ * Example:
+ *
+ *   Dir.chdir('/var/spool/mail')
+ *   Dir.pwd # => "/var/spool/mail"
+ *   dir  = Dir.new('/tmp')
+ *   fd = dir.fileno
+ *   Dir.fchdir(fd) do
+ *     Dir.pwd # => "/tmp"
+ *   end
+ *   Dir.pwd # => "/var/spool/mail"
+ *
+ * This method uses the
+ * {fchdir()}[https://www.man7.org/linux/man-pages/man3/fchdir.3p.html]
+ * function defined by POSIX 2008;
+ * the method is not implemented on non-POSIX platforms (raises NotImplementedError).
+ *
+ * Raises an exception if the file descriptor is not valid.
+ *
+ * In a multi-threaded program an error is raised if a thread attempts
+ * to open a +chdir+ block while another thread has one open,
+ * or a call to +chdir+ without a block occurs inside
+ * a block passed to +chdir+ (even in the same thread).
+ */
+static VALUE
+dir_s_fchdir(VALUE klass, VALUE fd_value)
+{
+    int fd = RB_NUM2INT(fd_value);
+
     if (chdir_blocking > 0) {
         if (rb_thread_current() != chdir_thread)
             rb_raise(rb_eRuntimeError, "conflicting chdir during another chdir block");
@@ -1079,23 +1301,58 @@ dir_s_chdir(int argc, VALUE *argv, VALUE obj)
     }
 
     if (rb_block_given_p()) {
-        struct chdir_data args;
-
-        args.old_path = rb_str_encode_ospath(rb_dir_getwd());
-        args.new_path = path;
+        struct fchdir_data args;
+        args.old_dir = dir_s_alloc(klass);
+        dir_initialize(NULL, args.old_dir, rb_fstring_cstr("."), Qnil);
+        args.fd = fd;
         args.done = FALSE;
-        return rb_ensure(chdir_yield, (VALUE)&args, chdir_restore, (VALUE)&args);
+        return rb_ensure(fchdir_yield, (VALUE)&args, fchdir_restore, (VALUE)&args);
     }
     else {
-        char *p = RSTRING_PTR(path);
-        int r = (int)(VALUE)rb_thread_call_without_gvl(nogvl_chdir, p,
-                                                        RUBY_UBF_IO, 0);
+        int r = (int)(VALUE)rb_thread_call_without_gvl(nogvl_fchdir, &fd,
+                                                       RUBY_UBF_IO, 0);
         if (r < 0)
-            rb_sys_fail_path(path);
+            rb_sys_fail("fchdir");
     }
 
     return INT2FIX(0);
 }
+#else
+#define dir_s_fchdir rb_f_notimplement
+#endif
+
+/*
+ * call-seq:
+ *   chdir -> 0
+ *   chdir { ... } -> object
+ *
+ * Changes the current working directory to +self+:
+ *
+ *   Dir.pwd # => "/"
+ *   dir = Dir.new('example')
+ *   dir.chdir
+ *   Dir.pwd # => "/example"
+ *
+ * With a block, temporarily changes the working directory:
+ *
+ * - Calls the block.
+ * - Changes to the given directory.
+ * - Executes the block (yields no args).
+ * - Restores the previous working directory.
+ * - Returns the block's return value.
+ *
+ * Uses Dir.fchdir if available, and Dir.chdir if not, see those
+ * methods for caveats.
+ */
+static VALUE
+dir_chdir(VALUE dir)
+{
+#if defined(HAVE_FCHDIR) && defined(HAVE_DIRFD) && HAVE_FCHDIR && HAVE_DIRFD
+    return dir_s_fchdir(rb_cDir, dir_fileno(dir));
+#else
+    return chdir_path(dir_get(dir)->path, false);
+#endif
+}
 
 #ifndef _WIN32
 VALUE
@@ -1143,16 +1400,14 @@ rb_dir_getwd(void)
 }
 
 /*
- *  call-seq:
- *     Dir.getwd -> string
- *     Dir.pwd -> string
+ * call-seq:
+ *   Dir.pwd -> string
  *
- *  Returns the path to the current working directory of this process as
- *  a string.
+ * Returns the path to the current working directory:
+ *
+ *   Dir.chdir("/tmp") # => 0
+ *   Dir.pwd           # => "/tmp"
  *
- *     Dir.chdir("/tmp")   #=> 0
- *     Dir.getwd           #=> "/tmp"
- *     Dir.pwd             #=> "/tmp"
  */
 static VALUE
 dir_s_getwd(VALUE dir)
@@ -1182,13 +1437,16 @@ check_dirname(VALUE dir)
 
 #if defined(HAVE_CHROOT)
 /*
- *  call-seq:
- *     Dir.chroot( string ) -> 0
+ * call-seq:
+ *   Dir.chroot(dirpath) -> 0
  *
- *  Changes this process's idea of the file system root. Only a
- *  privileged process may make this call. Not available on all
- *  platforms. On Unix systems, see <code>chroot(2)</code> for more
- *  information.
+ * Changes the root directory of the calling process to that specified in +dirpath+.
+ * The new root directory is used for pathnames beginning with <tt>'/'</tt>.
+ * The root directory is inherited by all children of the calling process.
+ *
+ * Only a privileged process may call +chroot+.
+ *
+ * See {Linux chroot}[https://man7.org/linux/man-pages/man2/chroot.2.html].
  */
 static VALUE
 dir_s_chroot(VALUE dir, VALUE path)
@@ -1217,18 +1475,20 @@ nogvl_mkdir(void *ptr)
 }
 
 /*
- *  call-seq:
- *     Dir.mkdir( string [, integer] ) -> 0
+ * call-seq:
+ *   Dir.mkdir(dirpath, permissions = 0775) -> 0
  *
- *  Makes a new directory named by <i>string</i>, with permissions
- *  specified by the optional parameter <i>anInteger</i>. The
- *  permissions may be modified by the value of File::umask, and are
- *  ignored on NT. Raises a SystemCallError if the directory cannot be
- *  created. See also the discussion of permissions in the class
- *  documentation for File.
+ * Creates a directory in the underlying file system
+ * at +dirpath+ with the given +permissions+;
+ * returns zero:
  *
- *    Dir.mkdir(File.join(Dir.home, ".foo"), 0700) #=> 0
+ *   Dir.mkdir('foo')
+ *   File.stat(Dir.new('foo')).mode.to_s(8)[1..4] # => "0755"
+ *   Dir.mkdir('bar', 0644)
+ *   File.stat(Dir.new('bar')).mode.to_s(8)[1..4] # => "0644"
  *
+ * See {File Permissions}[rdoc-ref:File@File+Permissions].
+ * Note that argument +permissions+ is ignored on Windows.
  */
 static VALUE
 dir_s_mkdir(int argc, VALUE *argv, VALUE obj)
@@ -1262,13 +1522,14 @@ nogvl_rmdir(void *ptr)
 }
 
 /*
- *  call-seq:
- *     Dir.delete( string ) -> 0
- *     Dir.rmdir( string ) -> 0
- *     Dir.unlink( string ) -> 0
+ * call-seq:
+ *   Dir.rmdir(dirpath) -> 0
  *
- *  Deletes the named directory. Raises a subclass of SystemCallError
- *  if the directory isn't empty.
+ * Removes the directory at +dirpath+ from the underlying file system:
+ *
+ *   Dir.rmdir('foo') # => 0
+ *
+ * Raises an exception if the directory is not empty.
  */
 static VALUE
 dir_s_rmdir(VALUE obj, VALUE dir)
@@ -1476,7 +1737,7 @@ nogvl_opendir_at(void *ptr)
             /* fallthrough*/
           case 0:
             if (fd >= 0) close(fd);
-            errno = e;
+            rb_errno_set(e);
         }
     }
 #else  /* !USE_OPENDIR_AT */
@@ -2869,7 +3130,7 @@ push_glob(VALUE ary, VALUE str, VALUE base, int flags)
     fd = AT_FDCWD;
     if (!NIL_P(base)) {
         if (!RB_TYPE_P(base, T_STRING) || !rb_enc_check(str, base)) {
-            struct dir_data *dirp = DATA_PTR(base);
+            struct dir_data *dirp = RTYPEDDATA_GET_DATA(base);
             if (!dirp->dir) dir_closed();
 #ifdef HAVE_DIRFD
             if ((fd = dirfd(dirp->dir)) == -1)
@@ -2993,26 +3254,35 @@ dir_open_dir(int argc, VALUE *argv)
 
 
 /*
- *  call-seq:
- *     Dir.foreach( dirname ) {| filename | block }                 -> nil
- *     Dir.foreach( dirname, encoding: enc ) {| filename | block }  -> nil
- *     Dir.foreach( dirname )                                       -> an_enumerator
- *     Dir.foreach( dirname, encoding: enc )                        -> an_enumerator
+ * call-seq:
+ *   Dir.foreach(dirpath, encoding: 'UTF-8') {|entry_name| ... }  -> nil
  *
- *  Calls the block once for each entry in the named directory, passing
- *  the filename of each entry as a parameter to the block.
+ * Calls the block with each entry name in the directory at +dirpath+;
+ * sets the given encoding onto each passed +entry_name+:
  *
- *  If no block is given, an enumerator is returned instead.
+ *   Dir.foreach('/example') {|entry_name| p entry_name }
  *
- *     Dir.foreach("testdir") {|x| puts "Got #{x}" }
+ * Output:
  *
- *  <em>produces:</em>
+ *   "config.h"
+ *   "lib"
+ *   "main.rb"
+ *   ".."
+ *   "."
  *
- *     Got .
- *     Got ..
- *     Got config.h
- *     Got main.rb
+ * Encoding:
  *
+ *   Dir.foreach('/example') {|entry_name| p entry_name.encoding; break }
+ *   Dir.foreach('/example', encoding: 'US-ASCII') {|entry_name| p entry_name.encoding; break }
+ *
+ * Output:
+ *
+ *   #<Encoding:UTF-8>
+ *   #<Encoding:US-ASCII>
+ *
+ * See {String Encoding}[rdoc-ref:encodings.rdoc@String+Encoding].
+ *
+ * Returns an enumerator if no block is given.
  */
 static VALUE
 dir_foreach(int argc, VALUE *argv, VALUE io)
@@ -3034,19 +3304,21 @@ dir_collect(VALUE dir)
 }
 
 /*
- *  call-seq:
- *     Dir.entries( dirname )                -> array
- *     Dir.entries( dirname, encoding: enc ) -> array
+ * call-seq:
+ *   Dir.entries(dirname, encoding: 'UTF-8') -> array
  *
- *  Returns an array containing all of the filenames in the given
- *  directory. Will raise a SystemCallError if the named directory
- *  doesn't exist.
+ * Returns an array of the entry names in the directory at +dirpath+;
+ * sets the given encoding onto each returned entry name:
  *
- *  The optional <i>encoding</i> keyword argument specifies the encoding of the
- *  directory. If not specified, the filesystem encoding is used.
+ *   Dir.entries('/example') # => ["config.h", "lib", "main.rb", "..", "."]
+ *   Dir.entries('/example').first.encoding
+ *   # => #<Encoding:UTF-8>
+ *   Dir.entries('/example', encoding: 'US-ASCII').first.encoding
+ *   # => #<Encoding:US-ASCII>
  *
- *     Dir.entries("testdir")   #=> [".", "..", "config.h", "main.rb"]
+ * See {String Encoding}[rdoc-ref:encodings.rdoc@String+Encoding].
  *
+ * Raises an exception if the directory does not exist.
  */
 static VALUE
 dir_entries(int argc, VALUE *argv, VALUE io)
@@ -3064,25 +3336,12 @@ dir_each_child(VALUE dir)
 }
 
 /*
- *  call-seq:
- *     Dir.each_child( dirname ) {| filename | block }                 -> nil
- *     Dir.each_child( dirname, encoding: enc ) {| filename | block }  -> nil
- *     Dir.each_child( dirname )                                       -> an_enumerator
- *     Dir.each_child( dirname, encoding: enc )                        -> an_enumerator
- *
- *  Calls the block once for each entry except for "." and ".." in the
- *  named directory, passing the filename of each entry as a parameter
- *  to the block.
- *
- *  If no block is given, an enumerator is returned instead.
- *
- *     Dir.each_child("testdir") {|x| puts "Got #{x}" }
- *
- *  <em>produces:</em>
- *
- *     Got config.h
- *     Got main.rb
+ * call-seq:
+ *   Dir.each_child(dirpath) {|entry_name| ... } -> nil
+ *   Dir.each_child(dirpath, encoding: 'UTF-8') {|entry_name| ... }  -> nil
  *
+ * Like Dir.foreach, except that entries <tt>'.'</tt> and <tt>'..'</tt>
+ * are not included.
  */
 static VALUE
 dir_s_each_child(int argc, VALUE *argv, VALUE io)
@@ -3096,24 +3355,22 @@ dir_s_each_child(int argc, VALUE *argv, VALUE io)
 }
 
 /*
- *  call-seq:
- *     dir.each_child {| filename | block }  -> dir
- *     dir.each_child                        -> an_enumerator
+ * call-seq:
+ *   each_child {|entry_name| ... } -> self
  *
- *  Calls the block once for each entry except for "." and ".." in
- *  this directory, passing the filename of each entry as a parameter
- *  to the block.
+ * Calls the block with each entry name in +self+
+ * except <tt>'.'</tt> and <tt>'..'</tt>:
  *
- *  If no block is given, an enumerator is returned instead.
+ *   dir = Dir.new('/example')
+ *   dir.each_child {|entry_name| p entry_name }
  *
- *     d = Dir.new("testdir")
- *     d.each_child  {|x| puts "Got #{x}" }
+ * Output:
  *
- *  <em>produces:</em>
- *
- *     Got config.h
- *     Got main.rb
+ *   "config.h"
+ *   "lib"
+ *   "main.rb"
  *
+ * If no block is given, returns an enumerator.
  */
 static VALUE
 dir_each_child_m(VALUE dir)
@@ -3123,14 +3380,14 @@ dir_each_child_m(VALUE dir)
 }
 
 /*
- *  call-seq:
- *     dir.children  -> array
+ * call-seq:
+ *   children -> array
  *
- *  Returns an array containing all of the filenames except for "."
- *  and ".." in this directory.
+ * Returns an array of the entry names in +self+
+ * except for <tt>'.'</tt> and <tt>'..'</tt>:
  *
- *     d = Dir.new("testdir")
- *     d.children   #=> ["config.h", "main.rb"]
+ *   dir = Dir.new('/example')
+ *   dir.children # => ["config.h", "lib", "main.rb"]
  *
  */
 static VALUE
@@ -3142,19 +3399,23 @@ dir_collect_children(VALUE dir)
 }
 
 /*
- *  call-seq:
- *     Dir.children( dirname )                -> array
- *     Dir.children( dirname, encoding: enc ) -> array
+ * call-seq:
+ *   Dir.children(dirpath) -> array
+ *   Dir.children(dirpath, encoding: 'UTF-8') -> array
  *
- *  Returns an array containing all of the filenames except for "."
- *  and ".." in the given directory. Will raise a SystemCallError if
- *  the named directory doesn't exist.
+ * Returns an array of the entry names in the directory at +dirpath+
+ * except for <tt>'.'</tt> and <tt>'..'</tt>;
+ * sets the given encoding onto each returned entry name:
  *
- *  The optional <i>encoding</i> keyword argument specifies the encoding of the
- *  directory. If not specified, the filesystem encoding is used.
+ *   Dir.children('/example') # => ["config.h", "lib", "main.rb"]
+ *   Dir.children('/example').first.encoding
+ *   # => #<Encoding:UTF-8>
+ *   Dir.children('/example', encoding: 'US-ASCII').first.encoding
+ *   # => #<Encoding:US-ASCII>
  *
- *     Dir.children("testdir")   #=> ["config.h", "main.rb"]
+ * See {String Encoding}[rdoc-ref:encodings.rdoc@String+Encoding].
  *
+ * Raises an exception if the directory does not exist.
  */
 static VALUE
 dir_s_children(int argc, VALUE *argv, VALUE io)
@@ -3228,12 +3489,16 @@ file_s_fnmatch(int argc, VALUE *argv, VALUE obj)
 }
 
 /*
- *  call-seq:
- *    Dir.home()       -> "/home/me"
- *    Dir.home("root") -> "/root"
+ * call-seq:
+ *   Dir.home(user_name = nil) -> dirpath
  *
- *  Returns the home directory of the current user or the named user
- *  if given.
+ * Retruns the home directory path of the user specified with +user_name+
+ * if it is not +nil+, or the current login user:
+ *
+ *   Dir.home         # => "/home/me"
+ *   Dir.home('root') # => "/root"
+ *
+ * Raises ArgumentError if +user_name+ is not a user name.
  */
 static VALUE
 dir_s_home(int argc, VALUE *argv, VALUE obj)
@@ -3258,10 +3523,15 @@ dir_s_home(int argc, VALUE *argv, VALUE obj)
 #if 0
 /*
  * call-seq:
- *   Dir.exist?(file_name)   ->  true or false
+ *   Dir.exist?(dirpath) ->  true or false
  *
- * Returns <code>true</code> if the named file is a directory,
- * <code>false</code> otherwise.
+ * Returns whether +dirpath+ is a directory in the underlying file system:
+ *
+ *   Dir.exist?('/example')         # => true
+ *   Dir.exist?('/nosuch')          # => false
+ *   Dir.exist?('/example/main.rb') # => false
+ *
+ * Same as File.directory?.
  *
  */
 VALUE
@@ -3288,8 +3558,7 @@ nogvl_dir_empty_p(void *ptr)
             /* fall through */
           case 0:
             if (e == ENOTDIR) return (void *)Qfalse;
-            errno = e; /* for rb_sys_fail_path */
-            return (void *)Qundef;
+            return (void *)INT2FIX(e);
         }
     }
     while ((dp = READDIR(dir, NULL)) != NULL) {
@@ -3304,10 +3573,18 @@ nogvl_dir_empty_p(void *ptr)
 
 /*
  * call-seq:
- *   Dir.empty?(path_name)  ->  true or false
+ *   Dir.empty?(dirpath) ->  true or false
  *
- * Returns <code>true</code> if the named file is an empty directory,
- * <code>false</code> if it is not a directory or non-empty.
+ * Returns whether +dirpath+ specifies an empty directory:
+ *
+ *   dirpath = '/tmp/foo'
+ *   Dir.mkdir(dirpath)
+ *   Dir.empty?(dirpath)            # => true
+ *   Dir.empty?('/example')         # => false
+ *   Dir.empty?('/example/main.rb') # => false
+ *
+ * Raises an exception if +dirpath+ does not specify a directory or file
+ * in the underlying file system.
  */
 static VALUE
 rb_dir_s_empty_p(VALUE obj, VALUE dirname)
@@ -3343,8 +3620,8 @@ rb_dir_s_empty_p(VALUE obj, VALUE dirname)
 
     result = (VALUE)rb_thread_call_without_gvl(nogvl_dir_empty_p, (void *)path,
                                             RUBY_UBF_IO, 0);
-    if (UNDEF_P(result)) {
-        rb_sys_fail_path(orig);
+    if (FIXNUM_P(result)) {
+        rb_syserr_fail_path((int)FIX2LONG(result), orig);
     }
     return result;
 }
@@ -3357,6 +3634,7 @@ Init_Dir(void)
     rb_include_module(rb_cDir, rb_mEnumerable);
 
     rb_define_alloc_func(rb_cDir, dir_s_alloc);
+    rb_define_singleton_method(rb_cDir,"for_fd", dir_s_for_fd, 1);
     rb_define_singleton_method(rb_cDir, "foreach", dir_foreach, -1);
     rb_define_singleton_method(rb_cDir, "entries", dir_entries, -1);
     rb_define_singleton_method(rb_cDir, "each_child", dir_s_each_child, -1);
@@ -3376,7 +3654,9 @@ Init_Dir(void)
     rb_define_method(rb_cDir,"pos", dir_tell, 0);
     rb_define_method(rb_cDir,"pos=", dir_set_pos, 1);
     rb_define_method(rb_cDir,"close", dir_close, 0);
+    rb_define_method(rb_cDir,"chdir", dir_chdir, 0);
 
+    rb_define_singleton_method(rb_cDir,"fchdir", dir_s_fchdir, 1);
     rb_define_singleton_method(rb_cDir,"chdir", dir_s_chdir, -1);
     rb_define_singleton_method(rb_cDir,"getwd", dir_s_getwd, 0);
     rb_define_singleton_method(rb_cDir,"pwd", dir_s_getwd, 0);
@@ -3393,51 +3673,26 @@ Init_Dir(void)
     rb_define_singleton_method(rb_cFile,"fnmatch", file_s_fnmatch, -1);
     rb_define_singleton_method(rb_cFile,"fnmatch?", file_s_fnmatch, -1);
 
-    /*  Document-const: File::Constants::FNM_NOESCAPE
-     *
-     *  Disables escapes in File.fnmatch and Dir.glob patterns
-     */
+    /* Document-const: FNM_NOESCAPE
+     * {File::FNM_NOESCAPE}[rdoc-ref:File::Constants@File-3A-3AFNM_NOESCAPE] */
     rb_file_const("FNM_NOESCAPE", INT2FIX(FNM_NOESCAPE));
-
-    /*  Document-const: File::Constants::FNM_PATHNAME
-     *
-     *  Wildcards in File.fnmatch and Dir.glob patterns do not match directory
-     *  separators
-     */
+    /* Document-const: FNM_PATHNAME
+     * {File::FNM_PATHNAME}[rdoc-ref:File::Constants@File-3A-3AFNM_PATHNAME] */
     rb_file_const("FNM_PATHNAME", INT2FIX(FNM_PATHNAME));
-
-    /*  Document-const: File::Constants::FNM_DOTMATCH
-     *
-     *  The '*' wildcard matches filenames starting with "." in File.fnmatch
-     *  and Dir.glob patterns
-     */
+    /* Document-const: FNM_DOTMATCH
+     * {File::FNM_DOTMATCH}[rdoc-ref:File::Constants@File-3A-3AFNM_DOTMATCH] */
     rb_file_const("FNM_DOTMATCH", INT2FIX(FNM_DOTMATCH));
-
-    /*  Document-const: File::Constants::FNM_CASEFOLD
-     *
-     *  Makes File.fnmatch patterns case insensitive (but not Dir.glob
-     *  patterns).
-     */
+    /* Document-const: FNM_CASEFOLD
+     * {File::FNM_CASEFOLD}[rdoc-ref:File::Constants@File-3A-3AFNM_CASEFOLD] */
     rb_file_const("FNM_CASEFOLD", INT2FIX(FNM_CASEFOLD));
-
-    /*  Document-const: File::Constants::FNM_EXTGLOB
-     *
-     *  Allows file globbing through "{a,b}" in File.fnmatch patterns.
-     */
+    /* Document-const: FNM_EXTGLOB
+     * {File::FNM_EXTGLOB}[rdoc-ref:File::Constants@File-3A-3AFNM_EXTGLOB] */
     rb_file_const("FNM_EXTGLOB", INT2FIX(FNM_EXTGLOB));
-
-    /*  Document-const: File::Constants::FNM_SYSCASE
-     *
-     *  System default case insensitiveness, equals to FNM_CASEFOLD or
-     *  0.
-     */
+    /* Document-const: FNM_SYSCASE
+     * {File::FNM_SYSCASE}[rdoc-ref:File::Constants@File-3A-3AFNM_SYSCASE] */
     rb_file_const("FNM_SYSCASE", INT2FIX(FNM_SYSCASE));
-
-    /*  Document-const: File::Constants::FNM_SHORTNAME
-     *
-     *  Makes patterns to match short names if existing.  Valid only
-     *  on Microsoft Windows.
-     */
+    /* Document-const: FNM_SHORTNAME
+     * {File::FNM_SHORTNAME}[rdoc-ref:File::Constants@File-3A-3AFNM_SHORTNAME] */
     rb_file_const("FNM_SHORTNAME", INT2FIX(FNM_SHORTNAME));
 }
 
