@@ -13,6 +13,10 @@
 extern int ruby_enable_coredump;
 int rb_get_next_signal(void);
 
+#ifdef POSIX_SIGNAL
+void (*ruby_posix_signal(int, void (*)(int)))(int);
+#endif
+
 RUBY_SYMBOL_EXPORT_BEGIN
 /* signal.c (export) */
 int rb_grantpt(int fd);
